from calliope._version import __version__

from calliope import examples, exceptions
from calliope.core import AttrDict, Model, read_netcdf
from calliope.core.util.logging import set_log_verbosity


__title__ = 'Calliope'
__author__ = 'Calliope contributors listed in AUTHORS'
__copyright__ = 'Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS'
